import asyncio
import time
from typing import TypeAlias

import discord
from v6d3music.core.mainasrc import main_for_raw_vc, raw_vc_for_member
from v6d3music.core.mainaudio import MainAudio

from v6d2ctx.context import Explicit

ResponseType: TypeAlias = list | dict | float | str | None


class Api:
    class MisusedApi(KeyError):
        def json(self) -> ResponseType:
            return {'error': list(map(str, self.args)), 'errormessage': str(self)}

    class UnknownApi(MisusedApi):
        pass

    class ExplicitFailure(MisusedApi):
        def __init__(self, explicit: Explicit) -> None:
            super().__init__(*explicit.args)

    def __init__(self, client: discord.Client, roles: dict[str, str]) -> None:
        self.client = client
        self.roles = roles

    def is_operator(self, user_id: int) -> bool:
        return '(operator)' in self.roles.get(f'roles{user_id}', '')

    async def api(self, request: dict, user_id: int) -> ResponseType:
        return await UserApi(self, request, user_id).api()


class UserApi:
    class UnknownMember(Api.MisusedApi):
        pass

    def __init__(self, api: Api, request: dict, user_id: int) -> None:
        self.pi = api
        self.client = api.client
        self.request = request
        self.user_id = user_id

    async def subs(self, requests: list[dict] | dict[str, dict]) -> ResponseType:
        match requests:
            case list():
                return list(
                    await asyncio.gather(
                        *(self.sub(request).api() for request in requests)
                    )
                )
            case dict():
                items = list(requests.items())
                responses = await asyncio.gather(
                    *(self.sub(request if 'type' in request else request | {'type': key}).api() for key, request in items)
                )
                return dict((key, response) for (key, _), response in zip(items, responses))
            case _:
                raise Api.MisusedApi('that should not happen')

    def sub(self, request: dict) -> 'UserApi':
        return UserApi(self.pi, request, self.user_id)

    async def _guild_api(self, guild_id: int) -> 'GuildApi':
        guild = self.client.get_guild(guild_id) or await self.client.fetch_guild(guild_id)
        if guild is None:
            raise UserApi.UnknownMember('unknown guild')
        member = guild.get_member(self.user_id) or await guild.fetch_member(self.user_id)
        if member is None:
            raise UserApi.UnknownMember('unknown member of a guild')
        return GuildApi(self.pi, self.request, member)

    async def _operator_api(self) -> 'OperatorApi':
        if not self.pi.is_operator(self.user_id):
            raise UserApi.UnknownMember('not an operator')
        return OperatorApi(self.pi, self.request, self.user_id)

    async def _api(self) -> ResponseType:
        match self.request:
            case {'guild': str() as guild_id_str} if guild_id_str.isdecimal() and len(guild_id_str) < 100:
                self.request.pop('guild')
                return await (await self._guild_api(int(guild_id_str))).api()
            case {'operator': _}:
                self.request.pop('operator')
                return await (await self._operator_api()).api()
            case {'type': 'ping', 't': (float() | int()) as t}:
                return time.time() - t
            case {'type': 'guilds'}:
                guilds = []
                for guild in self.client.guilds:
                    if guild.get_member(self.user_id) is not None:
                        guilds.append(str(guild.id))
                return guilds
            case {'type': '?'}:
                return 'this is user api'
            case {'type': '*', 'requests': list() | dict() as requests}:
                return await self.subs(requests)
            case _:
                raise Api.UnknownApi('unknown user api')

    async def api(self):
        try:
            try:
                return await self._api()
            except Explicit as e:
                raise Api.ExplicitFailure(e)
        except Api.MisusedApi as e:
            catches = self.request.get('catches', {})
            if len(e.args) and (key := e.args[0]) in catches:
                return catches[key]
            if '*' in catches:
                return e.json()
            raise


class GuildApi(UserApi):
    class VoiceNotConnected(Api.MisusedApi):
        pass

    def __init__(self, api: Api, request: dict, member: discord.Member) -> None:
        super().__init__(api, request, member.id)
        self.member = member
        self.guild = member.guild

    async def voice_api(self) -> 'VoiceApi':
        voice = self.member.voice
        if voice is None:
            raise GuildApi.VoiceNotConnected('you are not connected to voice')
        channel = voice.channel
        if channel is None:
            raise GuildApi.VoiceNotConnected('you are not connected to a voice channel')
        if self.client.user is None:
            raise GuildApi.VoiceNotConnected('bot client user not initialised')
        if self.client.user.id not in channel.voice_states:
            raise GuildApi.VoiceNotConnected('bot not connected')
        return VoiceApi(self.pi, self.request, self.member, channel)

    def sub(self, request: dict) -> 'GuildApi':
        return GuildApi(self.pi, request, self.member)

    async def _api(self) -> ResponseType:
        match self.request:
            case {'voice': _}:
                self.request.pop('voice')
                return await (await self.voice_api()).api()
            case {'type': '?'}:
                return 'this is guild api'
            case {'type': '*', 'requests': list() | dict() as requests}:
                return await self.subs(requests)
            case _:
                raise Api.UnknownApi('unknown guild api')


class VoiceApi(GuildApi):
    def __init__(
        self, api: Api, request: dict, member: discord.Member, channel: discord.VoiceChannel | discord.StageChannel
    ) -> None:
        super().__init__(api, request, member)
        self.channel = channel

    async def _main_api(self) -> 'MainApi':
        vc = await raw_vc_for_member(self.member)
        main = await main_for_raw_vc(vc, create=False, force_play=False)
        return MainApi(self.pi, self.request, self.member, self.channel, vc, main)

    def sub(self, request: dict) -> 'VoiceApi':
        return VoiceApi(self.pi, request, self.member, self.channel)

    async def _api(self) -> ResponseType:
        match self.request:
            case {'main': _}:
                self.request.pop('main')
                return await (await self._main_api()).api()
            case {'type': '?'}:
                return 'this is voice api'
            case {'type': '*', 'requests': list() | dict() as requests}:
                return await self.subs(requests)
            case _:
                raise Api.UnknownApi('unknown voice api')


class MainApi(VoiceApi):
    def __init__(
        self, api: Api, request: dict, member: discord.Member, channel: discord.VoiceChannel | discord.StageChannel,
        vc: discord.VoiceClient, main: MainAudio
    ) -> None:
        super().__init__(api, request, member, channel)
        self.vc = vc
        self.main = main

    def sub(self, request: dict) -> 'MainApi':
        return MainApi(self.pi, request, self.member, self.channel, self.vc, self.main)

    async def _api(self) -> ResponseType:
        match self.request:
            case {'type': 'volume'}:
                return self.main.volume
            case {'type': 'playing'}:
                return self.vc.is_playing()
            case {'type': 'queueformat'}:
                return await self.main.queue.format()
            case {'type': 'queuejson'}:
                return await self.main.queue.pubjson(self.member)
            case {'type': '?'}:
                return 'this is main api'
            case {'type': '*', 'requests': list() | dict() as requests}:
                return await self.subs(requests)
            case _:
                raise Api.UnknownApi('unknown main api')


class OperatorApi(UserApi):
    def sub(self, request: dict) -> 'OperatorApi':
        return OperatorApi(self.pi, request, self.user_id)

    async def _guild_visible(self, guild: discord.Guild) -> bool:
        return True

    async def _api(self) -> ResponseType:
        match self.request:
            case {'type': 'guilds'}:
                guilds = []
                for guild in self.client.guilds:
                    if self._guild_visible(guild):
                        guilds.append(
                            {
                                'id': str(guild.id),
                                'member_count': guild.member_count,
                                'name': guild.name,
                            }
                        )
                return guilds
            case {'type': '?'}:
                return 'this is operator api'
            case {'type': '*', 'requests': list() | dict() as requests}:
                return await self.subs(requests)
            case _:
                raise Api.UnknownApi('unknown operator api')
