import time
import contextlib
import asyncio
import os
import sys
import traceback

import discord
from v6d3music.app import MusicAppFactory, session_db
from v6d3music.commands import register_commands
from v6d3music.config import prefix
from v6d3music.core.cache_url import cache_db
from v6d3music.core.mainasrc import main_for_raw_vc, mainasrcs
from v6d3music.core.mainaudio import volume_db
from v6d3music.core.queueaudio import queue_db
from v6d3music.utils.entries_effects_for_args import effects_db

from rainbowadn.instrument import Instrumentation
from v6d1tokens.client import request_token
from v6d2ctx.handle_content import handle_content
from v6d2ctx.lock_for import lock_for
from v6d2ctx.pain import ABlockMonitor
from v6d2ctx.serve import serve

loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)
register_commands()


class MusicClient(discord.Client):
    async def close(self) -> None:
        save_task.cancel()
        await super().close()
        try:
            await save_all(False, False)
            print('saved')
        except Exception:
            traceback.print_exc()


client = MusicClient(
    intents=discord.Intents(
        members=True,
        guilds=True,
        bans=True,
        emojis=True,
        invites=True,
        voice_states=True,
        guild_messages=True,
        reactions=True,
        message_content=True,
    ),
    loop=loop
)

vcs_restored = False


async def restore_vcs():
    global vcs_restored
    vcs: list[tuple[int, int, bool]] = queue_db.get('vcs', [])
    try:
        for vcgid, vccid, vc_is_paused in vcs:
            try:
                guild: discord.Guild = await client.fetch_guild(vcgid)
                async with lock_for(guild, 'not in a guild'):
                    channels = await guild.fetch_channels()
                    channel: discord.VoiceChannel
                    channel, = [
                        ch for ch in
                        (
                            chc for chc in channels
                            if
                            isinstance(chc, discord.VoiceChannel)
                        )
                        if ch.id == vccid
                    ]
                    vp: discord.VoiceProtocol = await channel.connect()
                    assert isinstance(vp, discord.VoiceClient)
                    vc = vp
                    await main_for_raw_vc(vc, create=True, force_play=True)
                    if vc_is_paused:
                        vc.pause()
            except Exception as e:
                print(f'vc {vcgid} {vccid} {vc_is_paused} failed', e)
            else:
                print(f'vc restored {vcgid} {vccid}')
    finally:
        vcs_restored = True


@client.event
async def on_ready():
    print('ready')
    await client.change_presence(
        activity=discord.Game(
            name='феноменально',
        )
    )
    async with lock_for('vcs_restored', '...'):
        if not vcs_restored:
            await restore_vcs()


banned_guilds = set(map(int, map(str.strip, os.getenv('banned_guilds', '').split(':'))))


def guild_allowed(guild: discord.Guild | None) -> bool:
    return guild is not None and guild.id not in banned_guilds


def message_allowed(message: discord.Message) -> bool:
    return guild_allowed(message.guild)


@client.event
async def on_message(message: discord.Message) -> None:
    if message_allowed(message):
        await handle_content(message, message.content, prefix, client)


async def save_queues(delay: bool):
    for mainasrc in list(mainasrcs.values()):
        if delay:
            await asyncio.sleep(0.01)
        await mainasrc.queue.save(delay)


async def save_vcs(delay: bool):
    if vcs_restored:
        vcs = []
        vc: discord.VoiceClient
        for vc in (vcc for vcc in client.voice_clients if isinstance(vcc, discord.VoiceClient)):
            if delay:
                await asyncio.sleep(0.01)
            if vc.is_playing():
                if vc.guild is not None and vc.channel is not None:
                    vcs.append((vc.guild.id, vc.channel.id, vc.is_paused()))
        queue_db.set_nowait('vcs', vcs)


async def save_commit():
    await queue_db.commit()


async def save_all(delay: bool, save_playing: bool):
    await save_queues(delay)
    if save_playing:
        await save_vcs(delay)
    await save_commit()


async def save_job():
    while True:
        await asyncio.sleep(1)
        await save_all(True, True)


async def save_daemon():
    try:
        await save_job()
    except asyncio.CancelledError:
        pass


async def start_app():
    await MusicAppFactory.start(client)


async def setup_tasks():
    global save_task
    save_task = loop.create_task(save_daemon())
    loop.create_task(start_app())


class UpgradeABMInit(Instrumentation):
    def __init__(self):
        super().__init__(ABlockMonitor, '__init__')

    def instrument(self, method, abm, *, threshold=0.0, delta=10.0, interval=0.0):
        print('created upgraded')
        method(abm, threshold=threshold, delta=delta, interval=interval)
        abm.threshold = threshold


class UpgradeABMTask(Instrumentation):
    def __init__(self):
        super().__init__(ABlockMonitor, '_monitor')

    async def instrument(self, _, abm):
        print('started upgraded')
        while True:
            delta = abm.delta
            t = time.time()
            await asyncio.sleep(delta)
            spent = time.time() - t
            delay = spent - delta
            if delay > abm.threshold:
                abm.threshold = delay
                print(
                    f'upgraded block monitor reached new peak delay {delay:.4f}')
            interval = abm.interval
            if interval > 0:
                await asyncio.sleep(interval)


def _upgrade_abm() -> contextlib.ExitStack:
    with contextlib.ExitStack() as es:
        es.enter_context(UpgradeABMInit())
        es.enter_context(UpgradeABMTask())
        return es.pop_all()
    raise RuntimeError


async def main():
    async with volume_db, queue_db, cache_db, session_db, effects_db, ABlockMonitor(delta=0.5):
        if 'guerilla' in sys.argv:
            from pathlib import Path
            tokenpath = Path('.token.txt')
            if tokenpath.exists():
                token = tokenpath.read_text()
            else:
                token = input('token:')
                tokenpath.write_text(token)
        elif (token_ := os.getenv('trial_token')):
            token = token_
        else:
            token = await request_token('music', 'token')
        await client.login(token)
        loop.create_task(setup_tasks())
        if os.getenv('v6tor', None) is None:
            print('no tor')
        await client.connect()
        print('exited')


if __name__ == '__main__':
    with _upgrade_abm():
        serve(main(), client, loop)
